#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#include <doxml/xml.h>

struct {
  const char* descfile;
  const char* onlyQuote;
  const char* quotefile;
  int maxWidth;
} args={"desc.xml",
          "News flash! Butterflies exterminated in Sumatra: \"Stop hurricanes,\" vows Clinton",
          0,
          65};

typedef struct _Cell Cell;

typedef enum {
  Row,
  Column,
  Text
} CellKind;

struct _Cell {
  Cell* next;
  Cell* prev;
  CellKind kind;

  union {
    struct {
      Cell* firstChild;
      Cell* lastChild;
    } row,column;

    struct {
      const char* str;
      int canBreak;
    } text;
  } data;

  struct {
    int height,width;
    char** rows;
  } output;
};

Cell* buildCell(doxml_atom* desc);
void printCell(FILE* f,
               Cell* cell,
               int indent);
int assembleCell(Cell* cell,
                 int maxWidth);
const char* getQuote(void);

Cell* makeRowColumnCell(CellKind kind)
{
  Cell* res=(Cell*)malloc(sizeof(Cell));
  if (!res)
    {
      int tmp=errno;
      fprintf(stderr,
              "makeRowColumnCell(): malloc() failed: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  res->next=res->prev=0;
  res->kind=kind;
  res->data.row.firstChild=res->data.row.lastChild=0;
  return res;
}

Cell* makeRowCell(void)
{
  return makeRowColumnCell(Row);
}

Cell* makeColumnCell(void)
{
  return makeRowColumnCell(Column);
}

Cell* makeTextCell(const char* str,
                   int canBreak)
{
  Cell* res;

  if (!str)
    {
      fprintf(stderr,
              "makeTextCell(): str is NULL\n");
      errno=EFAULT;
      return 0;
    }

#ifdef DEBUG
  fprintf(stderr,"makeTextCell(\"%s\")\n",str);
#endif

  res=(Cell*)malloc(sizeof(Cell));
  if (!res)
    {
      int tmp=errno;
      fprintf(stderr,
              "makeTextCell(): malloc() failed: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  res->next=res->prev=0;
  res->kind=Text;
  res->data.text.str=strdup(str);
  if (!(res->data.text.str))
    {
      int tmp=errno;
      free(res);
      fprintf(stderr,
              "makeTextCell(): strdup() failed: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  res->data.text.canBreak=canBreak;

  return res;
}

int appendCell(Cell* parent,
               Cell* child)
{
  if (!parent)
    {
      fprintf(stderr,"appendCell(): parent is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (!child)
    {
      fprintf(stderr,"appendCell(): child is NULL\n");
      errno=EFAULT;
      return -1;
    }

  switch (parent->kind)
    {
    case Row:
    case Column:
      break;

    case Text:
      fprintf(stderr,"appendCell(): text cells cannot have children\n");
      errno=EINVAL;
      return -1;

    default:
      fprintf(stderr,"appendCell(): parent cell is of unknown type\n");
      errno=EINVAL;
      return -1;
    }

  if (parent->data.row.lastChild)
    parent->data.row.lastChild->next=child;
  child->prev=parent->data.row.lastChild;
  parent->data.row.lastChild=child;
  if (!(parent->data.row.firstChild))
    parent->data.row.firstChild=child;
  return 0;
}

int buildChildren(Cell* parent,
                  doxml_atom* desc)
{
  doxml_atom* cur;

  if (!parent)
    {
      fprintf(stderr,"buildChildren(): parent is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (!desc)
    {
      fprintf(stderr,"buildChildren(): desc is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (desc->type!=doxml_atom_element)
    {
      fprintf(stderr,"buildChildren(): desc is not an element\n");
      errno=EINVAL;
      return -1;
    }

  cur=desc->data.element.atoms.first;
  while (cur)
    {
      Cell* child=buildCell(cur);
      if (!child)
        {
          int tmp=errno;
          fprintf(stderr,"buildChildren(): buildCell() failed: %s\n",
                  strerror(tmp));
          errno=tmp;
          return -1;
        }

      if (appendCell(parent,child)<0)
        {
          int tmp=errno;
          fprintf(stderr,"buildChildren(): appendCell() failed: %s\n",
                  strerror(tmp));
          errno=tmp;
          return -1;
        }
          
      cur=cur->next;
    }

  return 0;
}

Cell* buildRowCell(doxml_atom* desc)
{
  Cell* res=makeRowCell();
  if (!res)
    {
      int tmp=errno;
      fprintf(stderr,"buildRowCell(): makeRowCell() failed: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  if (buildChildren(res,desc)<0)
    {
      int tmp=errno;
      fprintf(stderr,"buildRowCell(): buildChildren() failed: %s\n",
              strerror(tmp));
      errno=tmp;
      free(res);
      return 0;
    }

  return res;
}

Cell* buildColumnCell(doxml_atom* desc)
{
  Cell* res=makeColumnCell();
  if (!res)
    {
      int tmp=errno;
      fprintf(stderr,"buildColumnCell(): makeColumnCell() failed: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  if (buildChildren(res,desc)<0)
    {
      int tmp=errno;
      fprintf(stderr,"buildColumnCell(): buildChildren() failed: %s\n",
              strerror(tmp));
      errno=tmp;
      free(res);
      return 0;
    }

  return res;
}

const char* getText(doxml_atom* desc,
                    int* canBreak)
{
  static struct {
    char s[16384];
    int len;
  } buff;
  doxml_atom* cur;

  buff.s[0]=0;
  buff.len=0;
  (*canBreak)=0;

  cur=desc->data.element.atoms.first;

  while (cur)
    {
      const char* cur_str=0;

      switch (cur->type)
        {
        case doxml_atom_text:
          cur_str=cur->data.text.text;
          break;

        case doxml_atom_element:
          if (!strcmp(cur->data.element.name.name,"br"))
            cur_str="\n";
          else
            {
              if (!strcmp(cur->data.element.name.name,"quote"))
                {
                  cur_str=getQuote();
                  (*canBreak)=1;
                }
              else
                {
                  fprintf(stderr,
                          "getText(): unrecognized markup tag: \"%s\"\n",
                          cur->data.element.name.name);
                  errno=EINVAL;
                  return 0;
                }
            }
          break;

        case doxml_atom_pi:
          break;

        case doxml_atom_external_entity:
          fprintf(stderr,"getText(): external entities not supported\n");
          errno=EINVAL;
          return 0;
        }

      if (cur_str)
        {
          int len=strlen(cur_str);
          int new_len=buff.len+len;
          if ((new_len+1)>sizeof(buff.s))
            {
              fprintf(stderr,"getText(): text too long; buffer overflow\n");
              errno=ENOBUFS;
              return 0;
            }

          strcpy(buff.s+buff.len,cur_str);
          buff.len=new_len;
        }

      cur=cur->next;
    }

  return buff.s;
}

Cell* buildTextCell(doxml_atom* desc)
{
  int canBreak;
  const char* str=getText(desc,&canBreak);
  Cell* res=makeTextCell(str,canBreak);
  if (!res)
    {
      int tmp=errno;
      fprintf(stderr,"buildRowCell(): makeRowCell() failed: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  return res;
}

Cell* buildCell(doxml_atom* desc)
{
  Cell* res;

  if (!desc)
    {
      fprintf(stderr,"buildCell(): desc is NULL\n");
      errno=EFAULT;
      return 0;
    }

  if ((desc->type)!=doxml_atom_element)
    {
      fprintf(stderr,"buildCell(): desc is not an element\n");
      errno=EINVAL;
      return 0;
    }

  if (!strcmp(desc->data.element.name.name,"row"))
    return buildRowCell(desc);

  if (!strcmp(desc->data.element.name.name,"column"))
    return buildColumnCell(desc);

  if (!strcmp(desc->data.element.name.name,"text"))
    return buildTextCell(desc);

  fprintf(stderr,"buildCell(): unrecognized element name \"%s\"\n",
          desc->data.element.name.name);
  errno=EINVAL;
  return 0;
}

Cell* buildSignatureCell(doxml_atom* desc)
{
  if (!desc)
    {
      fprintf(stderr,"buildSignatureCell(): desc is NULL\n");
      errno=EFAULT;
      return 0;
    }

  if (desc->type!=doxml_atom_element)
    {
      fprintf(stderr,"buildSignatureCell(): desc is not an element\n");
      errno=EINVAL;
      return 0;
    }

  if (strcmp(desc->data.element.name.name,"signature"))
    {
      fprintf(stderr,"buildSignatureCell(): desc is not <signature>\n");
      errno=EINVAL;
      return 0;
    }

  return buildCell(desc->data.element.atoms.first);
}

void printSpaces(FILE* f,
                 int N)
{
  static char spaces[]="                    ";
  static int numSpaces=sizeof(spaces)-1;

  while (N>0)
    {
      if (N>numSpaces)
        {
          fprintf(f,"%s",spaces);
          N-=numSpaces;
        }
      else
        {
          fprintf(f,"%s",spaces+(numSpaces-N));
          N=0;
        }
    }
}

void printChildren(FILE* f,
                   Cell* parent,
                   int indent)
{
  Cell* cur=parent->data.row.firstChild;
  while (cur)
    {
      printCell(f,cur,indent);
      cur=cur->next;
    }
}

void printRowColumnCell(FILE* f,
                        Cell* cell,
                        int indent,
                        const char* typename)
{
  printSpaces(f,indent);
  fprintf(f,"{%s:\n",typename);
  printChildren(f,cell,indent+2);
  printSpaces(f,indent);
  fprintf(f," }\n");
}

void printRowCell(FILE* f,
                  Cell* cell,
                  int indent)
{
  printRowColumnCell(f,cell,indent,"row");
}

void printColumnCell(FILE* f,
                     Cell* cell,
                     int indent)
{
  printRowColumnCell(f,cell,indent,"column");
}

void printTextCell(FILE* f,
                   Cell* cell,
                   int indent)
{
  printSpaces(f,indent);
  fprintf(f,"{text:\n");
  fprintf(f,"%s\n",cell->data.text.str);
  printSpaces(f,indent);
  fprintf(f," }\n");
}

void printCell(FILE* f,
               Cell* cell,
               int indent)
{
  switch (cell->kind)
    {
    case Row:
      printRowCell(f,cell,indent);
      break;

    case Column:
      printColumnCell(f,cell,indent);
      break;

    case Text:
      printTextCell(f,cell,indent);
      break;

    default:
      fprintf(stderr,"printCell(): unknown cell type\n");
      return;
    }
}

int assembleChildren(Cell* cell,
                     int maxWidth,
                     int isRow)
{
  Cell* cur=cell->data.row.firstChild;
  while (cur)
    {
      if (assembleCell(cur,maxWidth)<0)
        {
          int tmp=errno;
          fprintf(stderr,"assembleChildren(): assembleCell() failed: %s\n",
                  strerror(tmp));
          errno=tmp;
          return -1;
        }
      if (isRow)
        maxWidth-=(cur->output.width);
      if ((cur=cur->next) && isRow)
        maxWidth--;
    }

  return 0;
}

int assembleRow(Cell* cell,
               int maxWidth)
{
  if (assembleChildren(cell,maxWidth,1)<0)
    {
      int tmp=errno;
      fprintf(stderr,"assembleRow(): assembleChildren() failed: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  {
    int width=0,height=0;
    Cell* cur=cell->data.row.firstChild;
    while (cur)
      {
        if ((cur->output.height)>height)
          height=cur->output.height;
        width+=cur->output.width;
        if (cur=cur->next)
          width++;
      }
    
    cell->output.width=width;
    cell->output.height=height;
  }

  cell->output.rows=(char**)malloc(sizeof(char*)*(cell->output.height));
  if (!(cell->output.rows))
    {
      int tmp=errno;
      fprintf(stderr,"assembleRow(): cannot malloc() rows: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  {
    int row;
    for (row=0; row<(cell->output.height); row++)
      {
        cell->output.rows[row]=(char*)malloc(cell->output.width+1);
        if (!(cell->output.rows[row]))
          {
            int tmp=errno;
            fprintf(stderr,"assembleRow(): cannot malloc() row[%d]: %s\n",
                    row,strerror(tmp));
            errno=tmp;
            return -1;
          }

        memset(cell->output.rows[row],' ',cell->output.width);
        cell->output.rows[row][cell->output.width]=0;

        {
          Cell* cur=cell->data.row.firstChild;
          int col=0;
          while (cur)
            {
              if (row<(cur->output.height))
                memcpy(cell->output.rows[row]+col,
                       cur->output.rows[row],
                       cur->output.width);
              col+=cur->output.width;
              if (cur=(cur->next))
                cell->output.rows[row][col++]='|';
            }
        }
      }
  }

  return 0;
}

int assembleColumn(Cell* cell,
                   int maxWidth)
{
  if (assembleChildren(cell,maxWidth,0)<0)
    {
      int tmp=errno;
      fprintf(stderr,"assembleColumn(): assembleChildren() failed: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  {
    int width=0,height=0;
    Cell* cur=cell->data.row.firstChild;
    while (cur)
      {
        if ((cur->output.width)>width)
          width=cur->output.width;
        height+=cur->output.height;
        if (cur=cur->next)
          height++;
      }
    
    cell->output.width=width;
    cell->output.height=height;
  }

  cell->output.rows=(char**)malloc(sizeof(char*)*(cell->output.height));
  if (!(cell->output.rows))
    {
      int tmp=errno;
      fprintf(stderr,"assembleColumn(): cannot malloc() rows: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  {
    int row;
    for (row=0; row<(cell->output.height); row++)
      {
        cell->output.rows[row]=(char*)malloc(cell->output.width+1);
        if (!(cell->output.rows[row]))
          {
            int tmp=errno;
            fprintf(stderr,"assembleColumn(): cannot malloc() row[%d]: %s\n",
                    row,strerror(tmp));
            errno=tmp;
            return -1;
          }

        memset(cell->output.rows[row],' ',cell->output.width);
        cell->output.rows[row][cell->output.width]=0;
      }
  }

  {
    Cell* cur=cell->data.row.firstChild;
    int row=0;
    while (cur)
      {
        int i;
        for (i=0; i<(cur->output.height); i++)
          memcpy(cell->output.rows[row++],
                 cur->output.rows[i],
                 cur->output.width);
        if (cur=(cur->next))
          memset(cell->output.rows[row++],'=',cell->output.width);
      }
  }

  return 0;
}

int assembleText(Cell* cell,
                 int maxWidth)
{
  typedef struct _Line Line;
  struct _Line {
    const char* start;
    int len;
    Line* next;
  };

  struct {
    Line* first;
    Line* last;
  } lines={0,0};

  {
    const char* cur=cell->data.text.str;
    while (cur)
      {
        const char* start=cur;
        const char* nl=strchr(cur,'\n');
        int len=0;

        if (nl)
          {
            len=nl-cur;
            cur=nl+1;
          }
        else
          {
            len=strlen(cur);
            cur=0;
          }

        {
          Line* l=(Line*)malloc(sizeof(Line));
          if (!l)
            {
              int tmp=errno;
              fprintf(stderr,"assembleText(): can't malloc() line: %s\n",
                      strerror(tmp));
              errno=tmp;
            }

          l->start=start;
          l->len=len;
          l->next=0;
          if (!lines.first)
            lines.first=l;
          if (lines.last)
            lines.last->next=l;
          lines.last=l;
        }
      }
  }

  if (cell->data.text.canBreak)
    {
      Line* l;
      for (l=lines.first; l; l=l->next)
        {
          if ((l->len)<=maxWidth)
            continue;

          {
            Line* rest;
            int breakPos=maxWidth;
            while ((!isspace(l->start[breakPos])) && (breakPos>0))
              breakPos--;

            rest=(Line*)malloc(sizeof(Line));
            if (!rest)
              {
                int tmp=errno;
                fprintf(stderr,"assembleText(): cannot malloc() rest\n",
                        strerror(tmp));
                errno=tmp;
                return -1;
              }

            rest->start=l->start+breakPos+1;
            rest->len=l->len-breakPos-1;
            l->len=breakPos;
            if (!(rest->next=l->next))
              lines.last=rest;
            l->next=rest;
          }
        }
    }

  {
    int width=0,height=0;
    Line* l;
    for (l=lines.first; l; l=l->next)
      {
        height++;
        if ((l->len)>width)
          width=l->len;
      }

    cell->output.width=width;
    cell->output.height=height;
  }

  cell->output.rows=(char**)malloc(sizeof(char*)*(cell->output.height));
  {
    Line* l=lines.first;
    int row;
    for (row=0; row<(cell->output.height); row++)
      {
        cell->output.rows[row]=(char*)malloc(cell->output.width+1);
        if (!(cell->output.rows[row]))
          {
            int tmp=errno;
            fprintf(stderr,"assembleText(): cannot malloc() row[%d]: %s\n",
                    row,strerror(tmp));
            errno=tmp;
            return -1;
          }

        memset(cell->output.rows[row],' ',cell->output.width);
        cell->output.rows[row][cell->output.width]=0;
        if (l)
          {
            memcpy(cell->output.rows[row],l->start,l->len);
            l=l->next;
          }
      }
  }

  {
    Line* l=lines.first;
    while (l)
      {
        Line* n=l->next;
        free(l);
        l=n;
      }
  }

  return 0;
}

int assembleCell(Cell* cell,
                 int maxWidth)
{
  if (!cell)
    {
      fprintf(stderr,"assembleCell(): cell is NULL\n");
      errno=EFAULT;
      return -1;
    }

  switch (cell->kind)
    {
    case Row:
      return assembleRow(cell,maxWidth);
      
    case Column:
      return assembleColumn(cell,maxWidth);

    case Text:
      return assembleText(cell,maxWidth);

    default:
      fprintf(stderr,"assembleCell(): unknown kind\n");
      errno=EINVAL;
      return -1;
    }
}

int renderCell(FILE* f,
               Cell* cell,
               int maxWidth)
{
  if (assembleCell(cell,maxWidth)<0)
    {
      int tmp=errno;
      fprintf(stderr,"renderCell(): assembleCell() failed: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  fputc('/',f);
  {
    int i;
    for (i=0; i<(cell->output.width); i++)
      fputc('=',f);
  }
  fputc('\\',f);
  fputc('\n',f);

  {
    int row;
    for (row=0; row<(cell->output.height); row++)
      fprintf(f,"|%s|\n",cell->output.rows[row]);
  }

  fputc('\\',f);
  {
    int i;
    for (i=0; i<(cell->output.width); i++)
      fputc('=',f);
  }
  fputc('/',f);
  fputc('\n',f);
}

const char* getQuote(void)
{
  doxml_document* doc;

  if (!args.quotefile)
    return args.onlyQuote;

  {
    doxml_context* context=doxml_open_file(args.quotefile);
    if (!context)
      {
        fprintf(stderr,"getQuote(): doxml_open_file() failed\n");
        return 0;
      }
    
    doc=doxml_parse(context);
    if (!doc)
      {
        fprintf(stderr,"getQuote(): doxml_parse() failed\n");
        return 0;
      }

    doxml_close(context);
  }

  {
    int N=0;
    doxml_atom* found=0;
    doxml_atom* cur;
    if (!(doc->root))
      {
        fprintf(stderr,"getQuote(): document has no root\n");
        return 0;
      }

    if ((doc->root->type)!=doxml_atom_element)
      {
        fprintf(stderr,"getQuote(): root atom is not an element.  Can't happen.\n");
        return 0;
      }

    for (cur=doc->root->data.element.atoms.first; cur; cur=cur->next)
      {
        if (!strcmp(cur->data.element.name.name,"quote"))
          {
            doxml_atom* text=cur->data.element.atoms.first;
            if (!text)
              continue;

            if ((text->type)!=doxml_atom_text)
              continue;

            if (text->next)
              continue;

            N++;

            if ((rand()%N)==0)
              found=text;
          }
      }

    if (!found)
      {
        fprintf(stderr,"getQuote(): no quote found (out of %d)\n",N);
        return 0;
      }

    return found->data.text.text;
  }
}

void usage(const char* progname)
{
  fprintf(stderr,"Usage:\n");
  fprintf(stderr,"%s [--desc <file>] [--quotes <file> | --quote <text>] [--maxWidth <N>]\n",progname);
  fprintf(stderr,"%s --help\n",progname);
}

int procargs(int argc, const char* argv[])
{
  int i;
  for (i=1; i<argc; i++)
    {
      if (!strcmp(argv[i],"--desc"))
        {
          args.descfile=argv[++i];
          if (!args.descfile)
            {
              fprintf(stderr,"--desc requires an argument\n");
              usage(argv[0]);
              return 10;
            }
        }

      if (!strcmp(argv[i],"--quotes"))
        {
          args.quotefile=argv[++i];
          if (!args.quotefile)
            {
              fprintf(stderr,"--quotes requires an argument\n");
              usage(argv[0]);
              return 12;
            }
        }

      if (!strcmp(argv[i],"--quote"))
        {
          args.onlyQuote=argv[++i];
          if (!args.onlyQuote)
            {
              fprintf(stderr,"--quote requires an argument\n");
              usage(argv[0]);
              return 13;
            }
        }

      if (!strcmp(argv[i],"--maxWidth"))
        {
          const char* maxWidthStr=argv[++i];
          if (!maxWidthStr)
            {
              fprintf(stderr,"--maxWidth requires an argument\n");
              usage(argv[0]);
              return 14;
            }

          {
            int w=atoi(maxWidthStr);
            if (w<=0)
              {
                fprintf(stderr,"Illegal value for --maxWidth: %s\n",
                        maxWidthStr);
                usage(argv[0]);
                return 15;
              }
            args.maxWidth=w;
          }
        }

      if (!strcmp(argv[i],"--help"))
        {
          usage(argv[0]);
          return 99;
        }
    }

  return 0;
}

int main(int argc, const char* argv[])
{
  doxml_document* doc;
  Cell* root_cell;

  {
    int err=procargs(argc,argv);
    if (err)
      return err;
  }

  srand(getpid());

  {
    doxml_context* context=doxml_open_file(args.descfile);
    if (!context)
      {
        fprintf(stderr,"doxml_open_file() failed\n");
        return 1;
      }
    
    doc=doxml_parse(context);
    if (!doc)
      {
        fprintf(stderr,"doxml_parse() failed\n");
        return 2;
      }

    doxml_close(context);
  }

  if (!(doc->root))
    {
      fprintf(stderr,"doc has no root element\n");
      return 3;
    }

  if (doc->root->type!=doxml_atom_element)
    {
      fprintf(stderr,"Root atom is not an element.  Can't happen\n");
      return 4;
    }

  root_cell=buildSignatureCell(doc->root);
  if (!root_cell)
    {
      int tmp=errno;
      fprintf(stderr,"buildSignatureCell() failed: %s\n",
              strerror(tmp));
      return 6;
    }

  renderCell(stdout,root_cell,args.maxWidth);
  return 0;
}
