/*
 * This is part of the doxml library, a parser for XML.  It should not
 *  require any functionality not defined in ANSI C's standard libraries.
 *
 * Copyright (C) 1999 John Stracke, http://www.thibault.org
 *
 * This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Log: delatom.c,v $
 * Revision 1.8  1999/07/19 01:31:11  francis
 * Implementing namespace reform (but not using it yet).  Not documented yet.
 *
 * Revision 1.7  1999/07/08 13:48:10  francis
 * Adding an (empty) clause for entity atoms, for clarity.  Adding a default,
 * as a sanity check.
 *
 * Revision 1.6  1999/06/07 04:57:38  francis
 * Implementing processing instructions.  Note the API changes.
 *
 * Revision 1.5  1999/06/01 01:51:14  francis
 * Adding functions to get/set/clear attributes.  Reorganzing the
 * to-do list.  Fixing a parser bug introduced I removed support for
 * attributes with no values at release 0.2.
 *
 * Revision 1.4  1999/05/21 23:10:28  francis
 * Reformatting, since renaming to doxml changed the widths of the
 * identifiers.
 *
 * Revision 1.3  1999/05/21 23:05:42  francis
 * Finished renaming to doxml
 *
 * Revision 1.2  1999/05/21 22:59:48  francis
 * Partially renaming to doxml
 *
 * Revision 1.1.1.1  1999/05/21 22:20:36  francis
 * Initial checkin of libxml, to be renamed doxml
 *
 */

static const char* id="$Id: delatom.c,v 1.8 1999/07/19 01:31:11 francis Exp $";

#include <doxml/xml.h>

static void
delete_name_contents(doxml_name* name)
{
  if (!name)
    return;

  if (name->space)
    free((char*)(name->space));
  if (name->name)
    free((char*)(name->name));
  if (name->prefix)
    free((char*)(name->prefix));
}

static void
delete_name_in_dtd_contents(doxml_name_in_dtd* name)
{
  if (!name)
    return;

  if (name->name)
    free((char*)(name->name));
  if (name->prefix)
    free((char*)(name->prefix));
}

void
doxml_delete_atom(doxml_atom* atom)
{
  if (!atom)
    return;

  switch (atom->type)
    {
    case doxml_atom_text:
      free((void*)(atom->data.text.text)); /* cast away const */
      break;

    case doxml_atom_element:
      delete_name_contents(&(atom->data.element.name));
      
      while (atom->data.element.attrs.first)
        doxml_attr_del(atom,atom->data.element.attrs.first);
      
      while (atom->data.element.atoms.first)
        doxml_delete_atom(atom->data.element.atoms.first);
      break;

    case doxml_atom_pi:
      delete_name_contents(&(atom->data.pi.name));
      if (atom->data.pi.value)
        free((void*)(atom->data.pi.value));
      break;

    case doxml_atom_external_entity:
      /* nothing to delete; decl just points to a declaration which
       *  is owned by the document's doxml_doctype.
       */
      break;

    default:
      fprintf(stderr,"doxml_delete_atom(): unknown atom type #%d\n",
              (int)(atom->type));
      break;
    }

  /* Update the parent's atom list */
  if (atom->next)
    atom->next->prev=atom->prev;
  else
    {
      if (atom->parent)
        atom->parent->data.element.atoms.last=atom->prev;
    }
  if (atom->prev)
    atom->prev->next=atom->next;
  else
    {
      if (atom->parent)
        atom->parent->data.element.atoms.first=atom->next;
    }
  
  free(atom);
}

void
doxml_delete_document(doxml_document* document)
{
  if (!document)
    return;

  if (document->xml_decl.version)
    free(document->xml_decl.version);
  if (document->xml_decl.encoding)
    free(document->xml_decl.encoding);

  if (document->prolog)
    {
      doxml_atom* cur;
      doxml_atom* next;
      for (cur=document->prolog; cur; cur=next)
        {
          next=cur->next;
          doxml_delete_atom(cur);
        }
    }

  if (document->root)
    doxml_delete_atom(document->root);

  free(document);
}

